import processing.core.*; 
import processing.xml.*; 

import controlP5.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class boids4 extends PApplet {

/**
 * Boids v4, 
 * Nicolas Liautaud 2009
 * <br/>
 * Boids flock simulation based on Craig Reynold's algorythm.
 * <br/>
 * Use BoidFlock, Obj and FpsCounter class
 */

ArrayList foodList, obstList;
int foodNbr, obstNbr, boidsNbr;
boolean displayAura, displayPerception;
boolean changeObstAura, changeFoodAura;
float obstAuraFactor, foodAuraFactor;
BoidsFlock boids;
FpsCounter fps;

public void setup() 
{ 
  // canvas
  size(600, 600);
  frameRate(30);
  smooth();

  // initialize obstacles
  obstList = new ArrayList();
  obstNbr  = 10;
  obstAuraFactor = 3;
  // initialize food
  foodList = new ArrayList();
  foodNbr  = 10;
  foodAuraFactor = 3;
  // initialize boids
  boids = new BoidsFlock();
  boidsNbr = 200;

  // initialize controls
  controlsInit();
  fps = new FpsCounter(width - 56, 42);

  // create foods
  for(int i=0; i < foodNbr; i++)
  {
    foodList.add(new Obj());
    Obj o = (Obj)foodList.get(i);
    o.aura = o.rad * foodAuraFactor;
  }
  // create obstacles
  for(int i=0; i < obstNbr; i++)
  {
    obstList.add(new Obj());
    Obj o = (Obj)obstList.get(i);
    o.aura = o.rad * obstAuraFactor;
  }
  // create boids
  for(int i=0; i < boidsNbr; i++)
  {
    boids.add();
  }
}

public void draw() 
{
  background(255);

  // draw details (controls manipulation)
  if(displayPerception) boids.displayPerception();
  if(displayAura) boids.displayAura();
  if(changeObstAura) changeObjAura(obstList, obstAuraFactor, 100, 100, 100);
  if(changeFoodAura) changeObjAura(foodList, foodAuraFactor, 100, 255, 100);

  // draw food
  for(int i=0; i<foodList.size(); i++)
  {
    Obj f = (Obj) foodList.get(i);
    f.display(120, 255, 120);
  }
  // draw obstacles
  for(int i=0; i<obstList.size(); i++)
  {
    Obj o = (Obj) obstList.get(i);
    o.display();
  }

  // draw boids and zone border
  boids.display();
  boids.zone.border(150);

  // hide control zone background and draw fps
  fill(255); 
  noStroke();
  rect(0, 0, width, 50);
  fps.display();

  // compute
  boids.update();

} 


public void mousePressed()
{
  // check if clic is in sim zone
  if(boids.zone.isIn(mouseX, mouseY))
  {
    // add food on left clic
    if(mouseButton == LEFT)
    {
      foodList.add(new Obj(mouseX, mouseY));
    }
    // add obst on right clic
    else if(mouseButton == RIGHT)
    {
      obstList.add(new Obj(mouseX, mouseY));
    }
    // delete things with middle clic
    else if(mouseButton == CENTER)
    {
      // delete food if mouse is on it
      for(int i=0; i<foodList.size(); i++){
        Obj o = (Obj) foodList.get(i);
        if(dist(o.pos.x, o.pos.y, mouseX, mouseY) < o.rad)
        {
          foodList.remove(o);
          break;
        }
      }
      // delete obstacle if mouse is on it
      for(int i=0; i<obstList.size(); i++){
        Obj o = (Obj) obstList.get(i);
        if(dist(o.pos.x, o.pos.y, mouseX, mouseY) < o.rad)
        {
          obstList.remove(o);
          break;
        }
      }
    }
  }
}

// stop displaying details on mouse releasing
public void mouseReleased()
{
  displayAura = false;
  displayPerception = false;
  changeFoodAura = false;
  changeObstAura = false;
}

// reset on SPACE press
public void keyPressed()
{
  if(key == ' ') setup();
}

// change aura value and display it (controls manipulation)
public void changeObjAura(ArrayList l, float f, int r, int g, int b)
{
  fill(r, g, b, 20); 
  stroke(r, g, b, 50);
  for(int i=0; i < l.size(); i++)
  {
    Obj o = (Obj) l.get(i);
    o.aura = o.rad * f;
    ellipse(o.pos.x,  o.pos.y, o.aura * 2, o.aura * 2);
  }
}




/**
 * Boid class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Used by BoidsFlock class
 */

class Boid
{
  PVector pos;
  PVector vel;

  /**
   * Class constructor.
   */
  Boid()
  {
    pos = boids.zone.randomPos();
    vel = new PVector(random(0,1), random(0,1));
  }
  /**
   * Class constructor specifying position of boid.
   */
  Boid(int x, int y)
  {
    init(x, y, random(0,1), random(0,1));
  }
  /**
   * Class constructor specifying position of boid.
   */
  Boid(PVector p)
  {
    init(PApplet.parseInt(p.x), PApplet.parseInt(p.y), random(0,1), random(0,1));
  }
  private void init(int px, int py, float vx, float vy)
  {
    pos = new PVector(px, py);
    vel = new PVector(vx, vy);
  }

  /**
   * <p>Boid cohesion : try to approach other boids.</p>
   * <p>Compute the average direction to the barycenter of the
   * percepted flock, and add a part of that to velocity.</p>
   */
  public void goToCenter()
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i < boids.size(); i++)
    {
      Boid b = (Boid) boids.get(i);
      if(b != this && PVector.dist(pos, b.pos) < boids.perception)
      {
        // add position of each other percepted boid
        v.add(b.pos);
        nbr++;
      }
    }
    if(nbr > 0)
    {
      v = PVector.sub(PVector.div(v, (float)nbr), pos); // do average
      v.mult(boids.cohesion * 0.001f); // set intensity
      vel.add(v); // and add to velocity
    }
  }

  /**
   * <p>Boid avoidance : try to keep a minimum distance between others.</p>
   * <p>Compute the average opposite way to the "touched"
   * flock, and add a part of that to velocity.</p>
   */
  public void keepDistance()
  {
    PVector u, v = new PVector();
    for(int i=0; i < boids.size(); i++)
    {
      Boid b = (Boid)boids.get(i);
      float d = PVector.dist(pos, b.pos);
      if(b != this && d < boids.aura)
      {
        u = PVector.sub(pos, b.pos);
        u.normalize();
        u.div(d);
        v.add(u);
      }
    }
    v.mult(boids.avoidance); // set intensity
    vel.add(v); // and add to velocity
  }

  /**
   * <p>Boid imitation : try to move in the same way than other boids.</p>
   * <p>Compute the average velocity of the percepted
   * flock, and add a part of that to velocity.</p>
   */
  public void matchVelocity()
  {
    PVector v = new PVector();
    int nbr = 0;
    for(int i=0; i < boids.size(); i++)
    { 
      Boid b = (Boid) boids.get(i);
      if(b != this && PVector.dist(pos, b.pos) < boids.perception)
      {
        // add velocity of each other percepted boid
        v.add(b.vel);
        nbr++;
      }
    }
    if(nbr > 0)
    {
      v = PVector.sub(PVector.div(v, nbr), vel); // do average
      v.mult(boids.imitation * 0.01f); // set intensity
      vel.add(v); // and add to velocity
    }
  }

  public void limitVelocity(float lim){
    if(vel.mag() > lim)
      vel = PVector.mult(PVector.div(vel, vel.mag()), lim);
  }

  /**
   * Keep a boid in zone by changing his velocity or position.
   *
   * @param  z Zone : zone to keep in
   * @param  loopZone boolean : loop or bounce on zone sides
   */
  public void keepInZone(Zone z, boolean loopZone)
  {
    int maxX = z.x + z.w;
    int maxY = z.y + z.h;
    if(!loopZone)
    {
      // When a boid approach a side, an opposite vector is added to velocity (bounce)
      float efficiency = .03f;
      // efficiency ~ 10 : boids immediately rejected
      // efficiency ~ .1  : boids slowly change direction
      PVector v = new PVector();
      if(pos.x <= z.x + boids.perception)       v.x = (z.x + boids.perception - pos.x) * efficiency;
      else if(pos.x >= maxX - boids.perception) v.x = (maxX - boids.perception - pos.x) * efficiency;
      if(pos.y <= z.y + boids.perception)       v.y = (z.y + boids.perception - pos.y) * efficiency;
      else if(pos.y >= maxY - boids.perception) v.y = (maxY - boids.perception - pos.y) * efficiency;
      vel.add(v);
    } 
    else
    {
      // When a boid cross a side, he's placed on the opposite side (teleport)
      if(pos.x <= z.x)       pos.x = maxX;
      else if(pos.x >= maxX) pos.x = z.x;
      if(pos.y <= z.y)       pos.y = maxY;
      else if(pos.y >= maxY) pos.y = z.y;
    }
  }

  /**
   * Affect boid velocity with an linear point force.
   *
   * @param  x x-coordinate of the effector
   * @param  y y-coordinate of the effector
   * @param  a aura of the effector (effect distance)
   * @param  i intensity of the effector. > 0 for attraction, < 0 for repulsion
   */
  public void effector(int x, int y, float a, float e)
  {
    float d = dist(x, y, pos.x, pos.y);
    if(d < a)
    {
      PVector v = new PVector();
      if(pos.x >= x) v.x += (a-d) * e * 0.01f;
      else v.x -= (a-d) * e * 0.01f;
      if(pos.y >= y) v.y += (a-d) * e * 0.01f;
      else v.y -= (a-d) * e * 0.01f;
      vel.sub(v);
    }
  }

  /**
   * Display boid.
   */
  public void display()
  {
    fill(210, 230, 250); 
    stroke(100,150,250);
    ellipse(pos.x, pos.y, boids.size, boids.size);
    float f = boids.size * 0.2f; // vector factor
    line(pos.x, pos.y, pos.x + vel.x * f, pos.y + vel.y * f);
  }
}
/**
 * BoidsFlock class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Use Boid and Zone class
 */

class BoidsFlock
{
  // Simulation zone
  Zone zone;
  // Keep in zone by looping or bouncing
  boolean loopZone;
  // Boids list
  ArrayList list;
  // Boids try to approach other boids (factor)
  float cohesion;
  // Boids try to keep a minimum distance between others (factor)
  float avoidance;
  // Boids try to move in the same way than other boids (factor)
  float imitation;
  // Boids radius
  int size = 7;
  // Distance for which boids touch things (avoidance, collisions)
  float aura;
  // Distance for which boids sees and responds to its environment
  float perception;
  // Define cruise-speed and stress-speed limit
  float speedLimit; 

  /**
   * Class constructor.
   */
  BoidsFlock()
  {
    zone = new Zone(0, 50, width, height-50);
    loopZone = true;
    list = new ArrayList();
    cohesion = 10;
    avoidance = 10;
    imitation = 10;
    size = 7;
    aura = 3 * size;
    perception = 10 * size;
    speedLimit = 4;
  }


  /**
   * Add a boid to flock.
   */
  public void add(){
    list.add(new Boid()); 
  }
  /**
   * Add a boid to flock at the defined position.
   */
  public void add(int x, int y){
    list.add(new Boid(x, y)); 
  }
  /**
   * Add a boid to flock at the defined position.
   */
  public void add(PVector p){
    list.add(new Boid(p)); 
  }

  /**
   * Add a boid to flock outside object list auras or don't create it.
   *
   * @param  objs Obj ArrayList : list of objects which areas should be avoided
   */
  public void addOutside(ArrayList objs)
  {
    boolean collide = false;
    int it = 0, itMax = 50000;
    PVector pos;
    // do this while random pos is in Obj zone
    do{
      it++;
      pos = boids.zone.randomPos();
      for(int i=0; i<objs.size(); i++)
      {
        Obj o = (Obj) objs.get(i);
        if(PVector.dist(pos, o.pos) < o.aura)
        {
          collide = true;
          break;
        }
      }
    }  
    while(collide && it < itMax);
    if(!collide) list.add(new Boid(pos)); 
  }

  /**
   * Delete boid from flock.
   *
   * @param  i boid index
   */
  public void del(int i){
    list.remove(i);
  }

  /**
   * Return the number of boids in flock.
   */
  public int size(){
    return list.size();
  }

  /**
   * Return the specified boid.
   *
   * @param i boid index
   * @return Boid
   */
  public Boid get(int i){
    return (Boid) list.get(i);
  }

  /**
   * Update each boid position and velocity according to simulation rules.
   */
  public void update()
  {
    for(int i=0; i<list.size(); i++)
    {
      Boid b = (Boid) list.get(i);

      // basic flock rules
      b.goToCenter();
      b.keepDistance();
      b.matchVelocity();

      // additionnal rules
      b.keepInZone(zone, loopZone);

      // Cruise-speed limitation
      b.limitVelocity(speedLimit);

      // food attraction and eat
      boolean contact;
      for(int j=0; j<foodList.size(); j++)
      { 
        Obj f = (Obj) foodList.get(j);
        // medium attraction when perceived
        b.effector(PApplet.parseInt(f.pos.x), PApplet.parseInt(f.pos.y), f.aura + boids.perception, 1.2f);
        // big repulsion on contact
        b.effector(PApplet.parseInt(f.pos.x), PApplet.parseInt(f.pos.y), f.rad + boids.size, -40);
        // if contact
        if(PVector.dist(f.pos, b.pos) < f.rad + boids.size)
        {
          f.rad -= .03f;// reduce obj size
          // if size is too small delete it and update control value
          if(f.rad <= 5){
            foodList.remove(j);
            Slider s = (Slider) controlP5.controller("food");
            s.setValue(foodList.size());
          }
        }
      }

      // obstacles boids.avoidance
      for(int j=0; j<obstList.size(); j++)
      { 
        Obj o = (Obj) obstList.get(j);
        // small repulsion when perceived (anticipation)
        b.effector(PApplet.parseInt(o.pos.x), PApplet.parseInt(o.pos.y), o.aura + boids.perception, -0.2f);
        // big repulsion on contact (collision)
        b.effector(PApplet.parseInt(o.pos.x), PApplet.parseInt(o.pos.y), o.rad + boids.size, -20);
      }

      // mouse repulsion
      b.effector(mouseX, mouseY, 70, -7);

      // Max-speed limitation
      //b.limitVelocity(speedLimit * 2);

      // finally update position
      b.pos.add(b.vel);
    }
  }

  /**
   * Draw each boid.
   */
  public void display()
  {
    for(int i=0; i < list.size(); i++)
    {
      Boid b = (Boid) list.get(i);
      b.display();
    } 
  }

  /**
   * Draw perception distance of boid alpha.
   */
  public void displayPerception()
  {
    fill(150, 180, 255, 50); 
    stroke(150, 180, 255);
    Boid b = (Boid) boids.get(0);
    ellipse(b.pos.x,  b.pos.y, boids.perception * 2, boids.perception * 2);
  }

  /**
   * Draw aura distance of boid alpha.
   */
  public void displayAura()
  {
    fill(255, 180, 150, 50); 
    stroke(255, 180, 150);
    Boid b = (Boid) boids.get(0);
    ellipse(b.pos.x,  b.pos.y, boids.aura * 2, boids.aura * 2);
  }
}

ControlP5 controlP5;
CheckBox checkbox;

public void controlsInit(){
  int x = 7;
  int y = 0;
  int w = PApplet.parseInt(width/7.9f);
  int h = 12;
  int e = 3;
  controlP5 = new ControlP5(this);

  controlP5.addSlider("boids", 0, 400, boidsNbr, x, e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("food", 0, 30, foodNbr, x, h+2*e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("obst", 0, 30, obstNbr, x, 2*h+3*e, PApplet.parseInt(w*1.2f), h);

  controlP5.addSlider("cohesion", 0, 50, boids.cohesion, w*2, 1 * e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("avoidance", 0, 50, boids.avoidance, w*2, 1 * h + 2 * e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("imitation", 0, 50, boids.imitation, w*2, 2 * h + 3 * e, PApplet.parseInt(w*1.2f), h);

  controlP5.addSlider("size", 1, 20, boids.size, w*4, e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("aura", 1, 100, boids.aura, w*4, h+2*e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("perception", 1, 200, boids.perception, w*4, 2*h+3*e, PApplet.parseInt(w*1.2f), h);

  controlP5.addSlider("food_aura", 1, 10, foodAuraFactor, w*6, e, PApplet.parseInt(w*1.2f), h);
  controlP5.addSlider("obst_aura", 1, 10, obstAuraFactor, w*6, h+2*e, PApplet.parseInt(w*1.2f), h);

  checkbox = controlP5.addCheckBox("canvasBounce", w*6, 2*h+3*e);  
  checkbox.setColorLabel(color(128));
  checkbox.addItem("Stay on canvas",0);

  String[] sliders = {
    "boids","food","obst", "cohesion","avoidance","imitation",
    "size","aura","perception", "food_aura", "obst_aura"  };
  for(int i=0;i<sliders.length;i++){
    Slider s = (Slider) controlP5.controller(sliders[i]);
    s.setColorLabel(color(0,0,0));
  }
}

public void controlEvent(ControlEvent theEvent) {
  if(theEvent.isGroup()) {
    // canvasBounce checkbox effect
    if(theEvent.group().name() == "canvasBounce"){
      int state = (int)theEvent.group().arrayValue()[0];
      if(state==1) boids.loopZone = false;
      else boids.loopZone = true;
    }
  }
}

// sliders effects

public void boids(int n) {
  // add boids
  if(n > boids.size()){
    while(boids.size()<n){
      boids.add();
    }
  }
  // remove boids
  else if(n<boids.size()){
    while(boids.size()>n){
      boids.del(0);
    } 
  }
}

public void food(int n) {
  // add food
  if(n>foodList.size()){
    while(foodList.size()<n){
      foodList.add(new Obj());
    }
  }
  // remove food
  else if(n<foodList.size()){
    while(foodList.size()>n){
      foodList.remove(0);
    } 
  }
}

public void obst(int n){
  // add obstacle
  if(n>obstList.size()){
    while(obstList.size()<n){
      obstList.add(new Obj());
    }
  }
  // remove obstacle
  else if(n<obstList.size()){
    while(obstList.size()>n){
      obstList.remove(0);
    } 
  }
}

public void cohesion(float n){
  boids.cohesion = n;
}
public void avoidance(float n){
  boids.avoidance = n;
}
public void imitation(float n){
  boids.imitation = n;
}

public void size(int n){
  boids.size = n;
  boids.aura = 3 * n;
  boids.perception = 10 * n;
}

public void aura(float n){
  boids.aura = n;
  displayAura = true;
}

public void perception(float n){
  boids.perception = n;
  displayPerception = true;
}

public void food_aura(float n){
  foodAuraFactor = n;
  changeFoodAura = true;
}

public void obst_aura(float n){
  obstAuraFactor = n;
  changeObstAura = true;
}



/**
 * FpsCounter class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Using exemple :
 * 
 * FpsCounter fps = new FpsCounter();
 * void draw(){
 * fps.display();
 * }
 */

class FpsCounter
{
  PVector position;
  float size;
  PFont font;
  int digits;
  private boolean change;
  private float fps, time;

  FpsCounter(){
    init(0,15);
  }
  FpsCounter(int x, int y){
    init(x,y);
  }
  private void init(int x, int y){
    change = true;
    position = new PVector(x, y);
    size = 10;
    digits = 1;
    font = createFont("Arial",1);
  }
  public void display(){
    time = millis();
    if(change && time%2 == 0){
      fps = roundTo(frameRate, digits);
      change = false;
    }
    else if(!change && time%2 != 0) change = true;
    display(fps);
  }
  private void display(float fps){
    textFont(font, size);
    fill(0);
    text("FPS " + fps, position.x, position.y);
  }
  public float roundTo(float a, int d){
    return round(a * pow(10,d)) / pow(10,d);
  }
}



/**
 * Obj class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Used for boids simulation
 */

class Obj
{
  PVector pos;
  float rad;
  float aura;

  Obj(){
    init(boids.zone.randomPos());
  }
  Obj(int x, int y){
    init(new PVector(x, y));
  }
  public void init(PVector p){
    pos = p;
    rad = random(10, 30);
    aura = 3 * rad;
  }

  public void display(){
    display(100, 100, 100);
  }
  public void display(int r, int g, int b){
    fill(r, g, b, 50);
    stroke(r, g, b);
    ellipse(pos.x, pos.y, rad * 2, rad * 2);
  }
}
/**
 * Zone class for Processing, 
 * Nicolas Liautaud 2009
 * <br/>
 * Used by BoidsFlock class
 */
class Zone
{
  int x, y, w, h;

  Zone(int _x, int _y, int _w, int _h){
    x = _x;
    y = _y;
    w = _w;
    h = _h;
  }

  // define if coord is in zone
  public boolean isIn(PVector v){
    return isIn(PApplet.parseInt(v.x), PApplet.parseInt(v.y));  
  }
  public boolean isIn(int px, int py){
    if(px > x && px < x + w &&
      py > y && py < y + h)
      return true;
    else return false;  
  }

  // return a random pos vector in zone
  public PVector randomPos(){
    return new PVector(random(x, x + w), random(y, y + h)); 
  }

  // display a border around zone
  public void border(int a){
    noFill(); 
    stroke(a);
    rect(x, y, w - 1, h - 1); 
  }
}




  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "boids4" });
  }
}
